import enum


def tiff_reader():
    # lazy load skimage
    import skimage.io

    return skimage.io.imread


def numpy_reader():
    # lazy load numpy
    import numpy as np

    return np.load


def tiff_writer():
    """
    Return a method that accepts (file, array) and saves it to the file.  File may be a file-like
    object, str, or pathlib.Path.
    """
    # lazy load skimage
    import skimage.io

    return lambda f, arr: skimage.io.imsave(f, arr, plugin="tifffile")


def numpy_writer():
    """
    Return a method that accepts (file, array) and saves it to the file.  File may be a file-like
    object, str, or pathlib.Path.
    """
    # lazy load numpy
    import numpy as np

    return np.save


class ImageFormat(enum.Enum):
    """
    The ImageFormat Enum exposes reading and writing methods for each enumerated object.

    To add a new object, assign to a name (e.g., NEW_FORMAT) a 4-tuple of (reader_provider,
    writer_provider, file_extension, {alternative_extensions}).
    """
    TIFF = (tiff_reader, tiff_writer, "tiff", {"tif"})
    NUMPY = (numpy_reader, numpy_writer, "npy", {})

    def __init__(
            self,
            reader_func,
            writer_func,
            file_ext,
            alternate_extensions,
    ):
        self._reader_func = reader_func
        self._writer_func = writer_func
        self._file_ext = file_ext
        self._alternate_extensions = set() if alternate_extensions is None else alternate_extensions

    @staticmethod
    def find_by_extension(extension):
        for imageformat in ImageFormat.__members__.values():
            if extension.lower() == imageformat._file_ext.lower():
                return imageformat
            for alternate_extension in imageformat._alternate_extensions:
                if extension.lower() == alternate_extension.lower():
                    return imageformat

        raise ValueError("Cannot find file format to match extension {}".format(extension))

    @property
    def reader_func(self):
        return self._reader_func()

    @property
    def writer_func(self):
        return self._writer_func()

    @property
    def file_ext(self):
        return self._file_ext



from __future__ import absolute_import, division, print_function, unicode_literals

import warnings

from ._typeformatting import format_tile_coordinates, format_tile_indices


class Tile(object):
    def __init__(self, coordinates, indices, tile_shape=None, sha256=None, extras=None):
        self.coordinates = format_tile_coordinates(coordinates)
        self.indices = format_tile_indices(indices)
        self._tile_shape = tuple(tile_shape) if tile_shape is not None else None
        self.sha256 = sha256
        self.extras = {} if extras is None else extras

        self._numpy_array = None
        self._numpy_array_future = None

    @property
    def tile_shape(self):
        if self._tile_shape is None:
            warnings.warn(
                "Decoding tile just to obtain shape.  It is recommended to include the tile shape "
                "in the tileset document to avoid this."
            )
            self._tile_shape = self._numpy_array_future().shape

        return self._tile_shape

    @property
    def numpy_array(self):
        if self._numpy_array is not None:
            return self._numpy_array
        else:
            result = self._numpy_array_future()

            if self._tile_shape is not None:
                assert self._tile_shape == result.shape
            self._tile_shape = result.shape

            return result

    @numpy_array.setter
    def numpy_array(self, numpy_array):
        if self._tile_shape is not None:
            assert self._tile_shape == numpy_array.shape

        self._numpy_array = numpy_array
        self._numpy_array_future = None
        self._tile_shape = self._numpy_array.shape

    def set_numpy_array_future(self, future):
        """
        Provides a tile with a callable, which should return the tile data when invoked.  It should
        be possible to invoke the callable 0, 1, or many times.

        Parameters
        ----------
        future : Callable[[], np.ndarray]
            A callable that yields the tile data when invoked.
        """
        self._numpy_array_future = future
        self._numpy_array = None

    def write(self, dst_fh, tile_format):
        """
        Write the contents of this tile out to a given file handle.
        """
        tile_format.writer_func(dst_fh, self.numpy_array)



from __future__ import absolute_import, division, print_function, unicode_literals

import os

from ._base import Backend, verify_checksum


class DiskBackend(Backend):
    def __init__(self, basedir):
        self._basedir = basedir

    def read_contextmanager(self, name, checksum_sha256=None):
        return _FileLikeContextManager(os.path.join(self._basedir, name), checksum_sha256)

    def write_file_handle(self, name=None):
        return open(os.path.join(self._basedir, name), "wb")


class _FileLikeContextManager(object):
    def __init__(self, path, checksum_sha256):
        self.path = path
        self.checksum_sha256 = checksum_sha256
        self.handle = None

    def __enter__(self):
        self.handle = open(self.path, "rb")
        verify_checksum(self.handle, self.checksum_sha256)
        return self.handle

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.handle is not None:
            return self.handle.__exit__(exc_type, exc_val, exc_tb)



from __future__ import absolute_import, division, print_function, unicode_literals

import requests
from io import BytesIO

from slicedimage.urlpath import pathjoin
from ._base import Backend, verify_checksum


class HttpBackend(Backend):
    def __init__(self, baseurl):
        self._baseurl = baseurl

    def read_contextmanager(self, name, checksum_sha256=None):
        parsed = pathjoin(self._baseurl, name)
        return _UrlContextManager(parsed, checksum_sha256)


class _UrlContextManager(object):
    def __init__(self, url, checksum_sha256):
        self.url = url
        self.checksum_sha256 = checksum_sha256
        self.handle = None

    def __enter__(self):
        resp = requests.get(self.url)
        resp.raise_for_status()
        self.handle = BytesIO(resp.content)
        verify_checksum(self.handle, self.checksum_sha256)
        return self.handle.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.handle.__exit__(exc_type, exc_val, exc_tb)



import shutil as _shutil
from tempfile import mkdtemp, template


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix=template, dir=None):
        self.name = mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        _shutil.rmtree(self.name)



import contextlib
import hashlib
import os
import sys
import tempfile
import time
import unittest

import requests
from requests import HTTPError

from slicedimage.backends import ChecksumValidationError, HttpBackend
from tests.utils import (
    ContextualChildProcess,
    TemporaryDirectory,
    unused_tcp_port,
)


class TestHttpBackend(unittest.TestCase):
    def setUp(self, timeout_seconds=5):
        self.contexts = []
        self.tempdir = TemporaryDirectory()
        self.contexts.append(self.tempdir)
        self.port = unused_tcp_port()

        if sys.version_info[0] == 2:
            module = "SimpleHTTPServer"
        elif sys.version_info[0] == 3:
            module = "http.server"
        else:
            raise Exception("unknown python version")

        self.contexts.append(ContextualChildProcess(
            [
                "python",
                "-m",
                module,
                str(self.port),
            ],
            cwd=self.tempdir.name,
        ).__enter__())

        end = time.time() + timeout_seconds
        while True:
            try:
                requests.get("http://0.0.0.0:{port}".format(port=self.port))
                break
            except requests.ConnectionError:
                if time.time() > end:
                    raise

        self.http_backend = HttpBackend("http://0.0.0.0:{port}".format(port=self.port))

    def tearDown(self):
        for context in self.contexts:
            context.__exit__(*sys.exc_info())

    def test_checksum_good(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            with self.http_backend.read_contextmanager(filename, expected_checksum) as cm:
                self.assertEqual(cm.read(), data)

    def test_checksum_bad(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            # make the hash incorrect
            expected_checksum = "{:x}".format(int(hashlib.sha256().hexdigest(), 16) + 1)

            with self.assertRaises(ChecksumValidationError):
                with self.http_backend.read_contextmanager(filename, expected_checksum) as cm:
                    self.assertEqual(cm.read(), data)

    def test_reentrant(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            with self.http_backend.read_contextmanager(filename, expected_checksum) as cm0:
                data0 = cm0.read(1)
                with self.http_backend.read_contextmanager(filename, expected_checksum) as cm1:
                    data1 = cm1.read()

                data0 = data0 + cm0.read()

                self.assertEqual(data, data0)
                self.assertEqual(data, data1)

    @staticmethod
    @contextlib.contextmanager
    def _test_checksum_setup(tempdir):
        """
        Write some random data to a temporary file and yield its path, the data, and the checksum of
        the data.
        """
        # write the file
        data = os.urandom(1024)

        expected_checksum = hashlib.sha256(data).hexdigest()

        with tempfile.NamedTemporaryFile(dir=tempdir) as tfh:
            tfh.write(data)
            tfh.flush()

            yield os.path.basename(tfh.name), data, expected_checksum

    def test_error(self):
        """
        Verifies that we raise an exception when we fail to find a file.
        """
        with self.assertRaises(HTTPError):
            backend = HttpBackend("http://0.0.0.0:{port}".format(port=self.port))
            with self.assertRaises(ChecksumValidationError):
                with backend.read_contextmanager("tileset.json") as cm:
                    cm.read()


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function, unicode_literals

import io
from threading import Lock

from diskcache import Cache

from ._base import Backend, verify_checksum

SIZE_LIMIT = 5e9
CACHE_VERSION = "v1"


class CachingBackend(Backend):
    _LOCK = Lock()
    _CACHE = {}

    def __init__(self, cacheroot, authoritative_backend, size_limit=SIZE_LIMIT):
        with CachingBackend._LOCK:
            if cacheroot not in CachingBackend._CACHE:
                CachingBackend._CACHE[cacheroot] = Cache(cacheroot, size_limit=int(size_limit))
            self._cache = CachingBackend._CACHE[cacheroot]
        self._authoritative_backend = authoritative_backend

    def read_contextmanager(self, name, checksum_sha256=None):
        if checksum_sha256 is not None:
            return _CachingBackendContextManager(
                self._authoritative_backend, self._cache, name, checksum_sha256)
        else:
            return self._authoritative_backend.read_contextmanager(
                name, checksum_sha256)

    def write_file_handle(self, name):
        return self._authoritative_backend.write_file_handle(name)


class _CachingBackendContextManager(object):
    def __init__(self, authoritative_backend, cache, name, checksum_sha256):
        self.authoritative_backend = authoritative_backend
        self.cache = cache
        self.name = name
        self.checksum_sha256 = checksum_sha256
        self.handle = None

    def __enter__(self):
        cache_key = "{}-{}".format(CACHE_VERSION, self.checksum_sha256)
        try:
            file_data = self.cache.read(cache_key)
        except KeyError:
            # not in cache :(
            with self.authoritative_backend.read_contextmanager(
                    self.name, self.checksum_sha256) as sfh:
                file_data = sfh.read()
            self.cache.set(cache_key, file_data)
            self.handle = io.BytesIO(file_data)

            # we are directly returning the data from the authoritative backend, which we assume
            # is verifying the checksum, so we don't layer on another checksum calculation.
        else:
            # If the data is small enough, the DiskCache library returns the cache data
            # as bytes instead of a buffered reader.
            # In that case, we want to wrap it in a file-like object.
            if isinstance(file_data, io.IOBase):
                self.handle = file_data
            else:
                self.handle = io.BytesIO(file_data)
            verify_checksum(self.handle, self.checksum_sha256)

        return self.handle.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.handle is not None:
            return self.handle.__exit__(exc_type, exc_val, exc_tb)



from __future__ import absolute_import, division, print_function, unicode_literals

import hashlib


class Backend(object):
    def read_contextmanager(self, name, checksum_sha256=None):
        """
        Returns a context manager, that when entered, should return a file-like object that can be
        read to get the data.

        It is possible to enter the context manager multiple times to yield the same data.

        If the checksum is provided and it does not match the checksum of the data read,
        ChecksumValidationError will be raised.

        Parameters
        ----------
        name : str
            The name of the file that is to be read.
        checksum_sha256 : Optional[str]
            The expected checksum of the file.
        """
        raise NotImplementedError()

    def write_file_handle(self, name):
        raise NotImplementedError()

    def write_file_from_handle(self, name, source_handle, block_size=(128 * 1024)):
        with self.write_file_handle(name) as dest_handle:
            data = source_handle.read(block_size)
            if len(data) == 0:
                return
            dest_handle.write(data)


class ChecksumValidationError(ValueError):
    """Raised when the downloaded file does not match the expected checksum."""
    pass


def verify_checksum(fh, expected_sha256_checksum, block_size=1024 * 1024):
    """
    Given a file-like handle, read from it in chunks to verify that the sha256 checksum matches
    `expected_sha256_checksum`.  If the checksum does not match, raise `ChecksumValidationError`.

    Before returning, the file handle is reset to the start of the file.
    :param fh: The file-like handle.
    :param expected_sha256_checksum: The expected sha256 checksum in hex.  If this parameter is
                                     None, then the method immediately returns.
    :param block_size: The block size of the IO.
    """
    if expected_sha256_checksum is None:
        return

    checksummer = hashlib.sha256()

    assert fh.tell() == 0
    while True:
        data = fh.read(block_size)
        checksummer.update(data)
        if len(data) == 0:
            calculated_checksum = checksummer.hexdigest()
            if calculated_checksum != expected_sha256_checksum:
                raise ChecksumValidationError(
                    "calculated checksum ({}) does not match expected checksum ({})".format(
                        calculated_checksum, expected_sha256_checksum))
            break

    fh.seek(0)



import subprocess


class ContextualChildProcess(object):
    """
    Provides a context manager for wrapping a child process.
    """
    def __init__(self, *args, **kwargs):
        self.proc = subprocess.Popen(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.proc.terminate()
        self.proc.wait()



from __future__ import absolute_import, division, print_function, unicode_literals

from ._typeformatting import format_tileset_dimensions, format_tileset_shape


class TileSet(object):
    def __init__(
            self,
            dimensions,
            shape,
            default_tile_shape=None,
            default_tile_format=None,
            extras=None):
        self.dimensions = format_tileset_dimensions(dimensions)
        self.shape = format_tileset_shape(shape)
        self.default_tile_shape = None if default_tile_shape is None else tuple(default_tile_shape)
        self.default_tile_format = default_tile_format
        self.extras = {} if extras is None else extras
        self._tiles = []

        self._discrete_dimensions = set()

    def __repr__(self):
        # get dimensions of optional shapes
        attributes = [
            "{k}: {v}".format(k=k, v=self.shape[k])
            for k in self.dimensions - {'y', 'x'}
            if k in self.shape
        ]
        xmin, xmax, ymin, ymax = float("inf"), float("-inf"), float("inf"), float("-inf")
        for tile in self._tiles:
            # try to get the shape in the following order:
            # 1. read the tile's declared shape without forcing a read & decode of the tile data.
            # 2. read the tileset's default tile shape.
            # 3. read the tile's shape through a read & decode.
            shape = tile._tile_shape
            if shape is None:
                shape = self.default_tile_shape
            if shape is None:
                shape = tile.tile_shape

            xmin = min(xmin, shape[0])
            xmax = max(xmax, shape[0])
            ymin = min(ymin, shape[1])
            ymax = max(ymax, shape[1])

        if xmin == xmax:
            attributes.append("x: {}".format(xmin))
        else:
            attributes.append("x: ({}-{})".format(xmin, xmax))
        if ymin == ymax:
            attributes.append("y: {}".format(ymin))
        else:
            attributes.append("y: ({}-{})".format(ymin, ymax))

        shape = ", ".join(attributes)
        return "<slicedimage.TileSet ({shape})>".format(shape=shape)

    def validate(self):
        raise NotImplementedError()

    def add_tile(self, tile):
        self._tiles.append(tile)

    def tiles(self, filter_fn=lambda _: True):
        """
        Return the tiles in this tileset.  If a filter_fn is provided, only the tiles for which
        filter_fn returns True are returned.
        """
        return list(filter(filter_fn, self._tiles))

    def get_dimension_shape(self, dimension_name):
        return self.shape[dimension_name]



import contextlib
import hashlib
import os
import tempfile
import unittest

from slicedimage.backends import ChecksumValidationError, DiskBackend


class TestDiskBackend(unittest.TestCase):
    def test_checksum_good(self):
        with self._test_checksum_setup() as setupdata:
            filepath, data, expected_checksum = setupdata

            backend = DiskBackend(os.path.dirname(filepath))
            with backend.read_contextmanager(os.path.basename(filepath), expected_checksum) as cm:
                self.assertEqual(cm.read(), data)

    def test_checksum_bad(self):
        with self._test_checksum_setup() as setupdata:
            filepath, data, expected_checksum = setupdata

            # make the hash incorrect
            expected_checksum = "{:x}".format(int(hashlib.sha256().hexdigest(), 16) + 1)

            backend = DiskBackend(os.path.dirname(filepath))
            with self.assertRaises(ChecksumValidationError):
                with backend.read_contextmanager(
                        os.path.basename(filepath),
                        expected_checksum) as cm:
                    self.assertEqual(cm.read(), data)

    def test_reentrant(self):
        with self._test_checksum_setup() as setupdata:
            filepath, data, expected_checksum = setupdata

            backend = DiskBackend(os.path.dirname(filepath))
            with backend.read_contextmanager(os.path.basename(filepath), expected_checksum) as cm0:
                data0 = cm0.read(1)
                with backend.read_contextmanager(
                        os.path.basename(filepath),
                        expected_checksum) as cm1:
                    data1 = cm1.read()

                data0 = data0 + cm0.read()

                self.assertEqual(data, data0)
                self.assertEqual(data, data1)

    @staticmethod
    @contextlib.contextmanager
    def _test_checksum_setup():
        """
        Write some random data to a temporary file and yield its path, the data, and the checksum of
        the data.
        """
        # write the file
        data = os.urandom(1024)

        expected_checksum = hashlib.sha256(data).hexdigest()

        with tempfile.NamedTemporaryFile() as tfh:
            tfh.write(data)
            tfh.flush()

            yield tfh.name, data, expected_checksum


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function, unicode_literals

import codecs
import hashlib
import json
import os
import tempfile
from io import BytesIO

from packaging import version
from six.moves import urllib

from slicedimage.urlpath import pathjoin, pathsplit
from .backends import CachingBackend, DiskBackend, HttpBackend, SIZE_LIMIT
from ._collection import Collection
from ._formats import ImageFormat
from ._tile import Tile
from ._tileset import TileSet


def infer_backend(baseurl, backend_config=None):
    """
    Guess the backend based on the format of `baseurl`, the consistent part of the URL or file path.
    The backend_config dictionary can contain flexible parameters for the different backends.

    Caching parameter keys include:

     - ["caching"]["directory"]  (default: None which disables caching)
     - ["caching"]["debug"]      (default: False)
     - ["caching"]["size_limit"] (default: SIZE_LIMIT)

    """

    parsed = urllib.parse.urlparse(baseurl)

    if parsed.scheme in ("http", "https"):
        backend = HttpBackend(baseurl)

        cache_config = {}
        if isinstance(backend_config, dict):
            cache_config = backend_config.get("caching", cache_config)

        cache_dir = cache_config.get("directory", None)
        if cache_dir is not None:
            cache_dir = os.path.expanduser(cache_dir)
            size_limit = cache_config.get("size_limit", SIZE_LIMIT)
            if size_limit > 0:
                debug = cache_config.get("debug", False)
                if debug:
                    print("> caching {} to {} (size_limit: {})".format(
                        baseurl, cache_dir, size_limit))
                backend = CachingBackend(cache_dir, backend, size_limit)
    elif parsed.scheme == "file":
        backend = DiskBackend(parsed.path)
    else:
        raise ValueError(
            "Unable to infer backend for url {}, please verify that baseurl points to a valid "
            "directory or web address".format(baseurl))
    return backend


def resolve_path_or_url(path_or_url, backend_config=None):
    """
    Given either a path (absolute or relative), or a URL, attempt to resolve it.  Returns a tuple
    consisting of: a :py:class:`slicedimage.backends._base.Backend`, the basename of the object, and
    the baseurl of the object.
    """
    try:
        return resolve_url(path_or_url, backend_config=None)
    except ValueError:
        if os.path.isfile(path_or_url):
            return resolve_url(
                os.path.basename(path_or_url),
                baseurl="file://{}".format(os.path.dirname(os.path.abspath(path_or_url))),
                backend_config=backend_config
            )
        raise


def _resolve_absolute_url(absolute_url, backend_config):
    """
    Given a string that is an absolute URL, return a tuple consisting of: a
    :py:class:`slicedimage.backends._base.Backend`, the basename of the object, and the baseurl of
    the object.
    """
    splitted = pathsplit(absolute_url)
    backend = infer_backend(splitted[0], backend_config)
    return backend, splitted[1], splitted[0]


def resolve_url(name_or_url, baseurl=None, backend_config=None):
    """
    Given a string that can either be a name or a fully qualified url, return a tuple consisting of:
    a :py:class:`slicedimage.backends._base.Backend`, the basename of the object, and the baseurl of
    the object.

    If the string is a name and not a fully qualified url, then baseurl must be set.  If the string
    is a fully qualified url, then baseurl is ignored.
    """
    try:
        # assume it's a fully qualified url.
        return _resolve_absolute_url(name_or_url, backend_config)
    except ValueError:
        if baseurl is None:
            # oh, we have no baseurl.  punt.
            raise
        absolute_url = pathjoin(baseurl, name_or_url)
        return _resolve_absolute_url(absolute_url, backend_config)


class Reader(object):
    @staticmethod
    def parse_doc(name_or_url, baseurl, backend_config=None):
        backend, name, baseurl = resolve_url(name_or_url, baseurl, backend_config)
        with backend.read_contextmanager(name) as fh:
            reader = codecs.getreader("utf-8")
            json_doc = json.load(reader(fh))
        doc_version = version.parse(json_doc[CommonPartitionKeys.VERSION])

        try:
            if doc_version >= version.parse(v0_0_0.VERSION):
                parser = v0_0_0.Reader()
            else:
                raise ValueError("Unrecognized version number")
        except KeyError:
            raise KeyError(
                "JSON document missing `version` field. Please specify the file format version.")

        return parser.parse(json_doc, baseurl, backend_config)

    def parse(self, json_doc, baseurl, backend_config):
        raise NotImplementedError()


class Writer(object):
    @staticmethod
    def write_to_path(partition, path, pretty=False, *args, **kwargs):
        document = v0_0_0.Writer().generate_partition_document(
            partition, path, pretty, *args, **kwargs)
        indent = 4 if pretty else None
        with open(path, "w") as fh:
            json.dump(document, fh, indent=indent, sort_keys=pretty)

    @staticmethod
    def default_partition_path_generator(parent_partition_path, partition_name):
        parent_partition_stem = os.path.splitext(os.path.basename(parent_partition_path))[0]
        partition_file = tempfile.NamedTemporaryFile(
            suffix=".json",
            prefix="{}-".format(parent_partition_stem),
            dir=os.path.dirname(parent_partition_path),
            delete=False,
        )
        return partition_file.name

    @staticmethod
    def default_tile_opener(tileset_path, tile, ext):
        tile_stemp = os.path.splitext(os.path.basename(tileset_path))[0]
        return tempfile.NamedTemporaryFile(
            suffix=".{}".format(ext),
            prefix="{}-".format(tile_stemp),
            dir=os.path.dirname(tileset_path),
            delete=False,
        )

    def generate_partition_document(self, partition, path, pretty=False, *args, **kwargs):
        raise NotImplementedError()


class v0_0_0(object):
    VERSION = "0.0.0"

    class Reader(Reader):
        def parse(self, json_doc, baseurl, backend_config):
            if CollectionKeys.CONTENTS in json_doc:
                # this is a Collection
                result = Collection(json_doc.get(CommonPartitionKeys.EXTRAS, None))
                for name, relative_path_or_url in json_doc[CollectionKeys.CONTENTS].items():
                    collection = Reader.parse_doc(relative_path_or_url, baseurl, backend_config)
                    collection._name_or_url = relative_path_or_url
                    result.add_partition(name, collection)
            elif TileSetKeys.TILES in json_doc:
                imageformat = json_doc.get(TileSetKeys.DEFAULT_TILE_FORMAT, None)
                if imageformat is not None:
                    imageformat = ImageFormat[imageformat]
                result = TileSet(
                    tuple(json_doc[TileSetKeys.DIMENSIONS]),
                    json_doc[TileSetKeys.SHAPE],
                    json_doc.get(TileSetKeys.DEFAULT_TILE_SHAPE, None),
                    imageformat,
                    json_doc.get(TileSetKeys.EXTRAS, None),
                )

                for tile_doc in json_doc[TileSetKeys.TILES]:
                    relative_path_or_url = tile_doc[TileKeys.FILE]
                    backend, name, _ = resolve_url(relative_path_or_url, baseurl, backend_config)

                    tile_format_str = tile_doc.get(TileKeys.TILE_FORMAT, None)
                    if tile_format_str:
                        tile_format = ImageFormat[tile_format_str]
                    else:
                        tile_format = result.default_tile_format
                    if tile_format is None:
                        # Still none :(
                        extension = os.path.splitext(name)[1].lstrip(".")
                        tile_format = ImageFormat.find_by_extension(extension)
                    checksum = tile_doc.get(TileKeys.SHA256, None)
                    tile = Tile(
                        tile_doc[TileKeys.COORDINATES],
                        tile_doc[TileKeys.INDICES],
                        tile_shape=tile_doc.get(TileKeys.TILE_SHAPE, None),
                        sha256=checksum,
                        extras=tile_doc.get(TileKeys.EXTRAS, None),
                    )

                    def future_maker(_source_fh_contextmanager, _tile_format):
                        """Produces a future that reads from a file and decodes according to the
                        specified file format."""
                        def _actual_future():
                            with _source_fh_contextmanager as fh:
                                return _tile_format.reader_func(fh)

                        return _actual_future

                    tile.set_numpy_array_future(
                        future_maker(
                            backend.read_contextmanager(name, checksum_sha256=checksum),
                            tile_format))
                    result.add_tile(tile)
            else:
                raise ValueError(
                    "JSON doc does not appear to be a collection partition or a tileset "
                    "partition. JSON doc must contain either a {contents} field pointing to a "
                    "tile manifest, or it must contain a {tiles} field that specifies a set of "
                    "tiles.".format(
                        contents=CollectionKeys.CONTENTS, tiles=TileSetKeys.TILES))

            return result

    class Writer(Writer):
        def generate_partition_document(
                self,
                partition,
                path,
                pretty=False,
                partition_path_generator=Writer.default_partition_path_generator,
                tile_opener=Writer.default_tile_opener,
                tile_format=ImageFormat.NUMPY,
        ):
            json_doc = {
                CommonPartitionKeys.VERSION: v0_0_0.VERSION,
                CommonPartitionKeys.EXTRAS: partition.extras,
            }
            if isinstance(partition, Collection):
                json_doc[CollectionKeys.CONTENTS] = dict()
                for partition_name, partition in partition._partitions.items():
                    partition_path = partition_path_generator(path, partition_name)
                    Writer.write_to_path(
                        partition, partition_path, pretty,
                        partition_path_generator=partition_path_generator,
                        tile_opener=tile_opener,
                        tile_format=tile_format,
                    )
                    json_doc[CollectionKeys.CONTENTS][partition_name] = os.path.basename(
                        partition_path)
                return json_doc
            elif isinstance(partition, TileSet):
                json_doc[TileSetKeys.DIMENSIONS] = tuple(partition.dimensions)
                json_doc[TileSetKeys.SHAPE] = partition.shape
                json_doc[TileSetKeys.TILES] = []

                if partition.default_tile_shape is not None:
                    json_doc[TileSetKeys.DEFAULT_TILE_SHAPE] = partition.default_tile_shape
                if partition.default_tile_format is not None:
                    json_doc[TileSetKeys.DEFAULT_TILE_FORMAT] = partition.default_tile_format.name
                if len(partition.extras) != 0:
                    json_doc[TileSetKeys.EXTRAS] = partition.extras

                for tile in partition._tiles:
                    tiledoc = {
                        TileKeys.COORDINATES: tile.coordinates,
                        TileKeys.INDICES: tile.indices,
                    }

                    with tile_opener(path, tile, tile_format.file_ext) as tile_fh:
                        buffer_fh = BytesIO()
                        tile.write(buffer_fh, tile_format)

                        buffer_fh.seek(0)
                        tile.sha256 = hashlib.sha256(buffer_fh.getvalue()).hexdigest()

                        buffer_fh.seek(0)
                        tile_fh.write(buffer_fh.read())
                        tiledoc[TileKeys.FILE] = os.path.basename(tile_fh.name)

                    if tile.tile_shape is not None:
                        tiledoc[TileKeys.TILE_SHAPE] = tile.tile_shape
                    tiledoc[TileKeys.SHA256] = tile.sha256
                    if tile_format is not None:
                        tiledoc[TileKeys.TILE_FORMAT] = tile_format.name
                    if len(tile.extras) != 0:
                        tiledoc[TileKeys.EXTRAS] = tile.extras
                    json_doc[TileSetKeys.TILES].append(tiledoc)

                return json_doc


class CommonPartitionKeys(object):
    VERSION = "version"
    EXTRAS = "extras"


class CollectionKeys(CommonPartitionKeys):
    CONTENTS = "contents"


class TileSetKeys(CommonPartitionKeys):
    DIMENSIONS = "dimensions"
    SHAPE = "shape"
    DEFAULT_TILE_SHAPE = "default_tile_shape"
    DEFAULT_TILE_FORMAT = "default_tile_format"
    TILES = "tiles"
    ZOOM = "zoom"


class TileKeys(object):
    FILE = "file"
    COORDINATES = "coordinates"
    INDICES = "indices"
    TILE_SHAPE = "tile_shape"
    TILE_FORMAT = "tile_format"
    SHA256 = "sha256"
    EXTRAS = "extras"



import os
import tempfile
import unittest
import uuid

from slicedimage.io import resolve_path_or_url, resolve_url


class TestResolvePathOrUrl(unittest.TestCase):
    def test_valid_local_path(self):
        with tempfile.NamedTemporaryFile() as tfn:
            abspath = os.path.realpath(tfn.name)
            _, name, baseurl = resolve_path_or_url(abspath)
            self.assertEqual(name, os.path.basename(abspath))
            self.assertEqual("file://{}".format(os.path.dirname(abspath)), baseurl)

            cwd = os.getcwd()
            try:
                os.chdir(os.path.dirname(abspath))
                _, name, baseurl = resolve_path_or_url(os.path.basename(abspath))
                self.assertEqual(name, os.path.basename(abspath))
                self.assertEqual("file://{}".format(os.path.dirname(abspath)), baseurl)
            finally:
                os.chdir(cwd)

    def test_invalid_local_path(self):
        with self.assertRaises(ValueError):
            resolve_path_or_url(str(uuid.uuid4()))

    def test_url(self):
        _, name, baseurl = resolve_path_or_url("https://github.com/abc/def")
        self.assertEqual(name, "def")
        self.assertEqual(baseurl, "https://github.com/abc")


class TestResolveUrl(unittest.TestCase):
    def test_fully_qualified_url(self):
        _, name, baseurl = resolve_url("https://github.com/abc/def")
        self.assertEqual(name, "def")
        self.assertEqual(baseurl, "https://github.com/abc")

        # even with a baseurl, this should work.
        _, name, baseurl = resolve_url("https://github.com/abc/def", "https://github.io")
        self.assertEqual(name, "def")
        self.assertEqual(baseurl, "https://github.com/abc")

    def test_relative_url(self):
        _, name, baseurl = resolve_url("def", "https://github.com/abc")
        self.assertEqual(name, "def")
        self.assertEqual(baseurl, "https://github.com/abc")

        # even with a path separator in the relative path, it should work.
        _, name, baseurl = resolve_url("abc/def", "https://github.com/")
        self.assertEqual(name, "def")
        self.assertEqual(baseurl, "https://github.com/abc")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import os
import setuptools

install_requires = [line.rstrip() for line in open(os.path.join(os.path.dirname(__file__), "requirements.txt"))]

setuptools.setup(
    name="slicedimage",
    version="1.0.4",
    description="Library to access sliced imaging data",
    author="Tony Tung",
    author_email="ttung@chanzuckerberg.com",
    license="MIT",
    packages=setuptools.find_packages(
        exclude=(
            "tests",
            "tests.*",
        )
    ),
    install_requires=install_requires,
    entry_points={
        'console_scripts': "slicedimage=slicedimage.cli.main:main"
    }
)



import contextlib
import hashlib
import os
import sys
import tempfile
import time
import unittest

import requests
from six import unichr

from slicedimage.backends import ChecksumValidationError, HttpBackend, CachingBackend
from tests.utils import (
    ContextualChildProcess,
    TemporaryDirectory,
    unused_tcp_port,
)


class TestCachingBackend(unittest.TestCase):
    def setUp(self, timeout_seconds=5):
        self.contexts = []
        self.tempdir = TemporaryDirectory()
        self.contexts.append(self.tempdir)
        self.port = unused_tcp_port()

        if sys.version_info[0] == 2:
            module = "SimpleHTTPServer"
        elif sys.version_info[0] == 3:
            module = "http.server"
        else:
            raise Exception("unknown python version")

        self.contexts.append(ContextualChildProcess(
            [
                "python",
                "-m",
                module,
                str(self.port),
            ],
            cwd=self.tempdir.name,
        ).__enter__())

        end = time.time() + timeout_seconds
        while True:
            try:
                requests.get("http://0.0.0.0:{port}".format(port=self.port))
                break
            except requests.ConnectionError:
                if time.time() > end:
                    raise

        self.cachedir = TemporaryDirectory()
        self.contexts.append(self.cachedir)

        self.http_backend = HttpBackend("http://0.0.0.0:{port}".format(port=self.port))
        self.caching_backend = CachingBackend(self.cachedir.name, self.http_backend)

    def tearDown(self):
        for context in self.contexts:
            context.__exit__(*sys.exc_info())

    def test_checksum_good(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm:
                self.assertEqual(cm.read(), data)

    def test_checksum_bad(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            # make the hash incorrect
            expected_checksum = "{:x}".format(int(hashlib.sha256().hexdigest(), 16) + 1)

            with self.assertRaises(ChecksumValidationError):
                with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm:
                    self.assertEqual(cm.read(), data)

    def test_cache_pollution(self):
        """
        Try to fetch a file but corrupt the data before fetching it.  The fetch should fail.

        Return the data to an uncorrupted state and try to fetch it again.  It should not have
        cached the bad data.
        """
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            # corrupt the file
            with open(os.path.join(self.tempdir.name, filename), "r+b") as fh:
                fh.seek(0)
                real_first_byte = fh.read(1).decode("latin-1")
                fh.seek(0)
                fh.write(unichr(ord(real_first_byte) ^ 0xff).encode("latin-1"))

            with self.assertRaises(ChecksumValidationError):
                with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm:
                    self.assertEqual(cm.read(), data)

            # un-corrupt the file
            with open(os.path.join(self.tempdir.name, filename), "r+b") as fh:
                fh.seek(0)
                real_first_byte = fh.read(1).decode("latin-1")
                fh.seek(0)
                fh.write(unichr(ord(real_first_byte) ^ 0xff).encode("latin-1"))

            with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm:
                self.assertEqual(cm.read(), data)

    def test_reentrant(self):
        with self._test_checksum_setup(self.tempdir.name) as setupdata:
            filename, data, expected_checksum = setupdata

            with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm0:
                data0 = cm0.read(1)
                with self.caching_backend.read_contextmanager(filename, expected_checksum) as cm1:
                    data1 = cm1.read()

                data0 = data0 + cm0.read()

                self.assertEqual(data, data0)
                self.assertEqual(data, data1)

    @staticmethod
    @contextlib.contextmanager
    def _test_checksum_setup(tempdir):
        """
        Write some random data to a temporary file and yield its path, the data, and the checksum of
        the data.
        """
        # write the file
        data = os.urandom(1024)

        expected_checksum = hashlib.sha256(data).hexdigest()

        with tempfile.NamedTemporaryFile(dir=tempdir) as tfh:
            tfh.write(data)
            tfh.flush()

            yield os.path.basename(tfh.name), data, expected_checksum


if __name__ == "__main__":
    unittest.main()



import collections
import json
import os
import unittest

import numpy as np
import skimage.io

import slicedimage
from tests.utils import build_skeleton_manifest, TemporaryDirectory

baseurl = "file://{}".format(os.path.abspath(os.path.dirname(__file__)))


class TestReader(unittest.TestCase):
    def test_read_tileset(self):
        result = slicedimage.Reader.parse_doc("tileset.json", baseurl)
        self.assertIsInstance(result, slicedimage.TileSet)
        self.assertEqual(result.shape, {'hyb': 4, 'ch': 4})
        self._verify_tiles(result.tiles())

    def test_read_collection(self):
        result = slicedimage.Reader.parse_doc("collection_l1.json", baseurl)
        self.assertIsInstance(result, slicedimage.Collection)
        tileset = result.find_tileset("fov_001")
        self.assertEqual(tileset.shape, {'hyb': 4, 'ch': 4})
        self._verify_tiles(result.tiles())

    def test_read_multilevel_collection(self):
        result = slicedimage.Reader.parse_doc("collection_l2.json", baseurl)
        self.assertIsInstance(result, slicedimage.Collection)
        tileset = result.find_tileset("fov_001")
        self.assertEqual(tileset.shape, {'hyb': 4, 'ch': 4})
        self._verify_tiles(result.tiles())

    def _verify_tiles(self, tile_generator):
        tiles = [tile for tile in tile_generator]
        self.assertEqual(len(tiles), 16)

        for hyb in range(4):
            for ch in range(4):
                for tile in tiles:
                    if tile.indices['hyb'] == hyb and tile.indices['ch'] == ch:
                        break
                else:
                    self.fail("Couldn't find tile of hyb {} ch {}".format(hyb, ch))

        for tile in tiles:
            for value in tile.coordinates.values():
                self.assertTrue(isinstance(value, collections.Hashable))
            for value in tile.indices.values():
                self.assertTrue(isinstance(value, collections.Hashable))


class TestFormats(unittest.TestCase):
    def test_tiff(self):
        """
        Generate a tileset consisting of a single TIFF tile, and then read it.
        """
        with TemporaryDirectory() as tempdir:
            # write the tiff file
            data = np.random.randint(0, 65535, size=(100, 100), dtype=np.uint16)
            skimage.io.imsave(os.path.join(tempdir, "tile.tiff"), data, plugin="tifffile")

            # TODO: (ttung) We should really be producing a tileset programmatically and writing it
            # disk.  However, our current write path only produces numpy output files.
            manifest = build_skeleton_manifest()
            manifest['tiles'].append(
                {
                    "coordinates": {
                        "x": [
                            0.0,
                            0.0001,
                        ],
                        "y": [
                            0.0,
                            0.0001,
                        ]
                    },
                    "indices": {
                        "hyb": 0,
                        "ch": 0,
                    },
                    "file": "tile.tiff",
                    "format": "tiff",
                },
            )
            with open(os.path.join(tempdir, "tileset.json"), "w") as fh:
                fh.write(json.dumps(manifest))

            result = slicedimage.Reader.parse_doc(
                "tileset.json",
                "file://{}".format(tempdir),
            )

            self.assertTrue(np.array_equal(list(result.tiles())[0].numpy_array, data))

    def test_numpy(self):
        """
        Generate a tileset consisting of a single TIFF tile, and then read it.
        """
        image = slicedimage.TileSet(
            ["x", "y", "ch", "hyb"],
            {'ch': 1, 'hyb': 1},
            (100, 100),
        )

        tile = slicedimage.Tile(
            {
                'x': (0.0, 0.01),
                'y': (0.0, 0.01),
            },
            {
                'hyb': 0,
                'ch': 0,
            },
        )
        tile.numpy_array = np.random.randint(0, 65535, size=(100, 100), dtype=np.uint16)
        image.add_tile(tile)

        with TemporaryDirectory() as tempdir:
            partition_path = os.path.join(tempdir, "tileset.json")
            partition_doc = slicedimage.v0_0_0.Writer().generate_partition_document(
                image, partition_path)
            with open(partition_path, "w") as fh:
                json.dump(partition_doc, fh)

            result = slicedimage.Reader.parse_doc(
                "tileset.json",
                "file://{}".format(tempdir),
            )

            self.assertTrue(np.array_equal(list(result.tiles())[0].numpy_array, tile.numpy_array))


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import, division, print_function, unicode_literals

import argparse

from . import checksum  # noqa
from ._base import CliCommand


def main():
    parser = argparse.ArgumentParser()

    subparser_root = parser.add_subparsers(dest="slicedimage_command_class")

    for cls in CliCommand.__subclasses__():
        subparser = cls.register_parser(subparser_root)
        subparser.set_defaults(slicedimage_command_class=(cls, subparser))

    args, argv = parser.parse_known_args()

    if args.slicedimage_command_class is None:
        parser.print_help()
        parser.exit(status=2)

    cls, subparser = args.slicedimage_command_class

    if len(argv) != 0:
        subparser.print_help()
        parser.exit(status=2)

    cls.run_command(args)



from __future__ import absolute_import, division, print_function, unicode_literals

from ._tileset import TileSet


class Collection(object):
    def __init__(self, extras=None):
        self.extras = extras
        self._partitions = dict()

    def validate(self):
        pass

    def add_partition(self, name, partition):
        self._partitions[name] = partition

    def all_tilesets(self):
        """
        Return all tilesets in this collection, directly or indirectly, as (name, tileset) tuples.
        """
        for name, partition in self._partitions.items():
            if isinstance(partition, Collection):
                for descendant_name, descendant_tileset in partition.all_tilesets():
                    yield descendant_name, descendant_tileset
            elif isinstance(partition, TileSet):
                yield name, partition

    def find_tileset(self, name):
        for partition_name, image_partition in self.all_tilesets():
            if name == partition_name:
                return image_partition
        return None

    def tiles(self, filter_fn=lambda _: True):
        result = []
        for partion_name, image_partition in self.all_tilesets():
            result.extend(image_partition.tiles(filter_fn))
        return result
